<?php

declare(strict_types=1);

namespace hood\vk\api\methods;

use Exception;

use hood\accounts\vk;
use hood\vk\robots\robot;
use hood\vk\api\attachments;

/**
 * Сообщение
 *
 * @method public static function put(RobotAbstract $from, int $to, string $message, int $mode = 2) Отправить сообщение
 *
 * @see https://vk.com/dev/messages.send
 * @see https://vk.com/dev/messages.getById
 *
 * @package hood\vk\api
 * @author Arsen Mirzaev Tatyano-Muradovich <red@hood.su>
 *
 * @todo Доработать строгий режим отправки: проверку сообщения в беседе (не имеет ID сообщений)
 */
final class messages extends method
{
    /**
     * @param int $mode Режим отправки
     */
    protected int $mode = 1;

    /**
     * @param string|null $text Текст
     */
    protected string|null $text = null;

    /**
     *  @var array Массив вложений
     */
    protected array $attachments;

    /**
     *  @var attachments Вложения
     */
    protected attachments $attachment;

    /**
     * Создать сообщение
     *
     * @param robot $robot Робот
     */
    public function __construct(
        protected robot $robot
    ) {
    }

    /**
     * Записать текст
     * 
     * @param string $text Текст
     * 
     * @return self
     */
    public function text(string $text): self
    {
        // Записать текст
        if (!isset($this->text)) {
            $this->text = $text;
        } else {
            $this->text .= $text;
        }

        return $this;
    }

    /**
     * Записать фото
     * 
     * @param $img Фото
     * 
     * @return self
     */
    public function image(...$imgs): self
    {
        // Перебор фото
        foreach ($imgs as $img) {

            // Загрузить фото
            $id = $this->robot->photo->getPhoto($img);

            // Записать к вложениям
            $this->attachment($id);
        }

        return $this;
    }

    /**
     * Записать вложение
     * 
     * @param $attachments Вложения
     */
    public function attachment(string ...$attachments): self
    {
        if (isset($this->attachment)) {
            // Если вложения инициализированны

            // Записать вложение
            $this->attachment->attachment(...$attachments);
        } else {
            // Если вложения не инициализированны

            // Инициализация вложений
            $this->attachment = new attachments($this->robot);

            // Записать вложение
            $this->attachment->attachment(...$attachments);
        }

        return $this;
    }

    /**
     * Отправить сообщение
     *
     * @param int|string|array $destination Получатель
     *
     * @see https://vk.com/dev/messages.send
     *
     * @return array Ответ сервера
     */
    public function send(int|string|array $destination): array
    {
        // Идентификатор сообщения
        $random_id = time();

        if ($this->mode = 1) {
            // Перемножение (по умолчанию)
            $random_id *= rand();
        }

        // Реиницилазиция
        $this->robot->api->reinit();

        // Цель отправки
        $this->robot->api->chooseDestination($destination);

        // Текст сообщения
        $this->robot->api['message'] = $this->text;

        // Идентификатор сообщения
        $this->robot->api['random_id'] = $random_id;

        // Фильтрация вложений, если они инициализированны
        if (isset($this->attachments)) {
            $forward_messages = [];
            foreach ($this->attachments as &$attachment) {
                if (iconv_substr($attachment, 0,  7, "UTF-8") === 'message') {

                    // Если среди вложений найдено сообщение для пересылки
                    $forward_messages[] = $attachment;
                    unset($attachment);
                }
            }

            if (!empty($forward_messages)) {
                // Если есть пересылаемые сообщения
                $this->robot->api['forward_messages'] = implode(',', $forward_messages);
            }

            if (!empty($this->attachments)) {
                // Если есть вложения
                $this->robot->api['attachment'] = implode(',', $this->attachments);
            }
        }


        // Запрос
        $request = $this->robot->browser->request(method: 'POST', uri: 'messages.send', options: ['form_params' => $this->robot->api['settings']]);

        // Очистка
        //unset($settings);

        if ($this->mode >= 2) {
            // Если установлен режим 2 (усиленная проверка отправленного сообщения)

            if (!empty($request["response"])) {
                // Если пришел ID сообщения

                // Реиницилазиция
                $this->robot->api->reinit();

                // Запрашиваемые сообщения
                $this->robot->api['message_ids'] = $request["response"];

                // Запрос
                if ($this->robot->browser->post(uri: 'https://api.vk.com/method/messages.getById', options: $this->robot->api['settings'])['response']['count'] === 0) {
                    // Если сообщения не существует, то повторить отправку
                    $this->send($destination);
                }
            } else {
                // Что-то придумать :)
            }
        }

        return (array) $request;
    }

    /**
     * Записать свойство
     *
     * @param string $name Название
     * @param mixed $value Значение
     *
     * @return void
     */
    public function __set(string $name, mixed $value): void
    {
        match ($name) {
            'attachment' => !isset($this->attachments) ? $this->attachments = $value : throw new Exception('Запрещено перезаписывать вложения'),
            'attachments' => !isset($this->attachments) ? $this->attachments = $value : throw new Exception('Запрещено перезаписывать массив вложений'),
            default => throw new Exception('Свойство не найдено: ' . $name)
        };
    }

    /**
     * Прочитать свойство
     *
     * @param string $name Название
     *
     * @return mixed
     */
    public function __get(string $name): mixed
    {
        return match ($name) {
            'text' => $this->text ?? throw new Exception('Текст не задан'),
            'attachment' => $this->attachment->attachments ??  $this->attachment = new attachments($this->robot),
            'attachments' => isset($this->attachment) ? $this->attachment->attachments : throw new Exception('Вложения не инициализированны'),
            default => throw new Exception('Свойство не найдено: ' . $name)
        };
    }

    /**
     * Получить информацию о сообщении
     *
     * @return array Информация о сообщении
     */
    public function info(): array
    {
        //Инициализация
        //$settings = $this->init();
        $this->robot->api->init();

        // Цель отправки
        $this->robot->api->chooseDestination($this->destination);

        // Сообщение
        $this->robot->api['message'] = $this->message;

        // Режим отправки
        $this->robot->api['mode'] = $this->mode;

        // Фильтрация вложений
        $forward_messages = [];
        foreach ($this->attachments as &$attachment) {
            if (iconv_substr(str: $attachment, offset: 0, length: 7, charset: "UTF-8") === 'message') {
                // Если среди вложений найдено сообщение для пересылки

                $forward_messages[] = $attachment;
                unset($attachment);
            }
        }

        if (!empty($forward_messages)) {
            // Если есть пересылаемые сообщения

            $this->robot->api['forward_messages'] = implode(',', $forward_messages);
        }

        if (!empty($attachments)) {
            // Если есть вложения

            $this->robot->api['attachment'] = implode(',', $this->attachments);
        }

        return $settings;
    }
}
